/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include "ServerCapFiller.h"
#include "base/Log.h"
#include "base/util/utils.h"
#include "http/constants.h"

#include "pim/ClientSettings.h"

USE_NAMESPACE

#define MAX_HTTP_BUFFER_SIZE    2048

static char* createHttpErrorMessage(DWORD errorCode) {

    WCHAR* errorMessage = new WCHAR[512];
    memset(errorMessage, 0, 512);

    FormatMessage(
                FORMAT_MESSAGE_FROM_HMODULE,
                GetModuleHandle(L"wininet.dll"),
                errorCode,
                MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT),
                errorMessage,
                512,
                NULL);

    if (!errorMessage || wcslen(errorMessage) == 0) {
        wsprintf(errorMessage, L"Unknown error.");
    }

    char* ret = toMultibyte(errorMessage);
    if (errorMessage) delete [] errorMessage;
    return ret;
}

const char* ServerCapFiller::getServerCap(){
    
    char* bufferA = new char[MAX_HTTP_BUFFER_SIZE+1];
    int status = -1;
    unsigned int contentLength = 0;
    WCHAR* wurlHost      = NULL;
    WCHAR* wurlResource  = NULL;
    char*  p             = NULL;
    char*  response      = NULL;
    bool error           = false;

	StringBuffer fullUrl(getRegConfig()->getAccessConfig().getSyncURL());
	fullUrl.append("/info");
	URL url(fullUrl.c_str());
    
    DWORD size  = 0,
          read  = 0,
          flags = INTERNET_FLAG_RELOAD |
                  INTERNET_FLAG_NO_CACHE_WRITE |
                  INTERNET_FLAG_KEEP_CONNECTION |           // This is necessary if authentication is required.
                  INTERNET_FLAG_NO_COOKIES;                 // This is used to avoid possible server errors on successive sessions.

	LPCWSTR acceptTypes[2] = {TEXT("*/*"), NULL};

    // Set flags for secure connection (https).
    if (url.isSecure()) {
        flags = flags
              | INTERNET_FLAG_SECURE
              | INTERNET_FLAG_IGNORE_CERT_CN_INVALID
              | INTERNET_FLAG_IGNORE_CERT_DATE_INVALID
              ;
    }

    HINTERNET hSession = InternetOpen(TEXT("WinInet Progress Sample"),
                                      INTERNET_OPEN_TYPE_PRECONFIG,
                                      NULL,
                                      NULL,
                                      0);
	if(!hSession){
        LOG.error("ServerCapFiller hSession is null: return");
        error = true;
		goto finally;
        
	}
	wurlResource = toWideChar(url.resource);
    wurlHost     = toWideChar(url.host);
    HINTERNET hConnection = InternetConnect(hSession, 
                                            wurlHost,  // Server
                                            url.port,
                                            NULL,     // Username
                                            NULL,     // Password
                                            INTERNET_SERVICE_HTTP,
                                            0,        // Synchronous
                                            NULL);    // No Context

	if(!hConnection){
        LOG.error("ServerCapFiller hConnection is null: return");
        error = true;
        goto finally;		
	}
    
    HINTERNET hRequest = HttpOpenRequest(hConnection, 
                                         TEXT("GET"),
                                         wurlResource,
                                         NULL,    // Default HTTP Version
                                         NULL,    // No Referer
                                         acceptTypes, // Accept anything
                                         0,       // Flags
                                         NULL);   // No Context


    if(!hRequest){
        LOG.error("ServerCapFiller hRequest is null: return");
        error = true;
        goto finally;		
	}

    HttpSendRequest(hRequest,
                    NULL,    // No extra headers
                    0,       // Header length
                    NULL,    // No Body
                    0);      // Body length
    
    size = sizeof(status);
    HttpQueryInfo (hRequest,
                   HTTP_QUERY_STATUS_CODE | HTTP_QUERY_FLAG_NUMBER,
                   (LPDWORD)&status,
                   (LPDWORD)&size,
                   NULL);
    
    if (status == HTTP_STATUS_OK) {
    	LOG.debug("Data sent succesfully to server. Server responds OK");        
    }
    
    //Initialize response
    contentLength = 0;
    HttpQueryInfo (hRequest,
                   HTTP_QUERY_CONTENT_LENGTH | HTTP_QUERY_FLAG_NUMBER,
                   (LPDWORD)&contentLength,
                   (LPDWORD)&size,
                   NULL);

    LOG.debug("%s", "ServerCapFiller: reading the http message for server capabilities");
    LOG.debug("Content-length: %u", contentLength);

    if (contentLength <= 0) {
        contentLength = 2048;
        LOG.debug("Undefined content-length = %u. Using the a fixed buffer of 2048", contentLength);        
    }
    
    // Allocate a block of memory for response read.
    response = new char[contentLength+1];
    if (response == NULL) {
        //lastErrorCode = ERR_NOT_ENOUGH_MEMORY;
        //sprintf(lastErrorMsg, "Not enough memory to allocate a buffer for the server response: %d required.", contentLength);
        setErrorF(ERR_NOT_ENOUGH_MEMORY, "Not enough memory to allocate a buffer for the server response: %d required.", contentLength);
        LOG.error("%s", getLastErrorMsg());
        error = true;
        goto finally;
    }
    memset(response, 0, contentLength);
	p = response;
    int realResponseLenght = 0;
        
    do {
        if (!InternetReadFile(hRequest, (LPVOID)bufferA, MAX_HTTP_BUFFER_SIZE, &read)) {
            DWORD code = GetLastError();
            //lastErrorCode = ERR_READING_CONTENT;
            char* tmp = createHttpErrorMessage(code);
            LOG.error("getCapFiller: http error getting the server info: %s", tmp);
            delete [] tmp;
            error = true;
            goto finally;
		}

        // Sanity check: some proxy could send additional bytes.
        // Correct 'read' value to be sure we won't overflow the 'response' buffer.
        if ((realResponseLenght + read) > contentLength) {
            LOG.debug("Warning! %d bytes read -> truncating data to content-lenght = %d.", (realResponseLenght + read), contentLength);
            read = contentLength - realResponseLenght;
        }

        if (read > 0) {
            memcpy(p, bufferA, read);               // Note: memcopy exactly the bytes read (could be no readable chars...)
            p += read;
            realResponseLenght += read;
        }

    } while (read);

    // free read buffer
    delete [] bufferA; bufferA = NULL;
    
    if (realResponseLenght <= 0) {
        LOG.error("getCapFiller: Error reading HTTP response from Server: received data of size = %d.", realResponseLenght);
        error = true;
        goto finally;
    }

    // Log bytes read if different from content length
    // (should be already the same...)
    if (realResponseLenght != contentLength) {
        LOG.info("Bytes read: ", realResponseLenght);
    	contentLength = realResponseLenght;
    }
    response[contentLength] = 0;

finally:

    // Close the Internet handles.
    if (hSession) {
        InternetCloseHandle (hSession);
    }
    if (hConnection) {
        InternetCloseHandle (hConnection);
    }
    if (hRequest) {
        InternetCloseHandle (hRequest);
    }   

    if (wurlHost)     delete [] wurlHost;
    if (wurlResource) delete [] wurlResource;
    if (bufferA)      delete [] bufferA;
    
    if (error) {
        delete [] response; response = NULL;
    }

    return response;
}

//Http request on the server url to see the server capabilities
//we store them in a DeviceConfig object
void ServerCapFiller::fillServerCap(){
    
    const char* cap = getServerCap();    
	StringBuffer msg(cap);
    delete [] cap;

	if(msg.empty()){
		LOG.error("No response from server while catching the server caps");
		getRegConfig()->getServerConfig().setSmartSlowSync(2);
	} else {

        msg.replaceAll("\n\n","\n");
        ArrayList vals; msg.split(vals, "\n");
        
        for(int i = 1; i<vals.size(); i++) {
            StringBuffer man(((StringBuffer*)vals[i])->c_str());
            if (man.find("Man=")==0){
                getRegConfig()->getServerConfig().setMan(man.substr(strlen("Man="), (man.length()-1)));
                continue;
            }
            if (man.find("Mod=")==0){
                getRegConfig()->getServerConfig().setMod(man.substr(strlen("Mod="), (man.length()-1)));
                continue;
            }
            if (man.find("SwV=")==0){
                getRegConfig()->getServerConfig().setSwv(man.substr(strlen("SwV="), (man.length()-1)));
                continue;
            }
            if (man.find("HwV=")==0){
                getRegConfig()->getServerConfig().setHwv(man.substr(strlen("HwV="), (man.length()-1)));
                continue;
            }
            if (man.find("FwV=")==0){
                getRegConfig()->getServerConfig().setFwv(man.substr(strlen("FwV="), (man.length()-1)));
                continue;
            }
            if (man.find("OEM=")==0){
                getRegConfig()->getServerConfig().setOem(man.substr(strlen("OEM="), (man.length()-1)));
                continue;
            }
            if (man.find("DevID=")==0){
                getRegConfig()->getServerConfig().setDevID(man.substr(strlen("DevID="), (man.length()-1)));
                continue;
            }
            if (man.find("DevTyp=")==0){
                getRegConfig()->getServerConfig().setDevType(man.substr(strlen("DevTyp="), (man.length()-1)));
                continue;
            }
            if (man.find("VerDTD=")==0){
                getRegConfig()->getServerConfig().setVerDTD((man.substr(strlen("VerDTD="), (man.length()-1))));
                continue;
            }
            if (man.find("UTC=")==0){
			    //man.substr(strlen("UTC="), (man.length()-1));
			    getRegConfig()->getServerConfig().setUtc((man.find("true")!= StringBuffer::npos)?true : false);
                continue;
            }
            if (man.find("SupportLargeObjs=")==0){
			    //man.substr(strlen("SupportLargeObjs="), (man.length()-1));
			    getRegConfig()->getServerConfig().setLoSupport((man.find("true")!=StringBuffer::npos)? true : false);
                continue;
            }
		    if (man.find("SupportNumberOfChanges=")==0){
			    //man.substr(strlen("SupportNumberOfChanges="), (man.length()-1));
			    getRegConfig()->getServerConfig().setNocSupport((man.find("true")!=StringBuffer::npos)? true : false);
                continue;
            }
            if (man.find("Ext=")==0){
                //Ext=X-funambol-smartslow. If the X-funambol-smartslow is found, set the 
                // SmartSlowSync to 0. Otherwise leave it to undefined (2) to be set at the end
                // of the loop.
                StringBuffer val(man.substr(strlen("Ext="), (man.length()-1)));
                if(val.find("X-funambol-smartslow") != StringBuffer::npos){
                    getRegConfig()->getServerConfig().setSmartSlowSync(0);
                } 
                continue;
            }
            
            /*
            * to handle old server properties
            */
            if (man.find("manufacturer=")==0){
                getRegConfig()->getServerConfig().setMan(man.substr(strlen("manufacturer="), (man.length()-1)));
                continue;
            }
            if (man.find("model=")==0){
                getRegConfig()->getServerConfig().setMod(man.substr(strlen("model="), (man.length()-1)));
                continue;
            }
            if (man.find("software version=")==0){
                getRegConfig()->getServerConfig().setSwv(man.substr(strlen("software version="), (man.length()-1)));
                continue;
            }
            if (man.find("hardware version=")==0){
                getRegConfig()->getServerConfig().setHwv(man.substr(strlen("hardware version="), (man.length()-1)));
                continue;
            }
            if (man.find("firmware version=")==0){
                getRegConfig()->getServerConfig().setFwv(man.substr(strlen("firmware version="), (man.length()-1)));
                continue;
            }
            if (man.find("device id=")==0){
                getRegConfig()->getServerConfig().setDevID(man.substr(strlen("device id="), (man.length()-1)));
                continue;
            }
            if (man.find("DTD version=")==0){
                getRegConfig()->getServerConfig().setVerDTD(man.substr(strlen("DTD version="), (man.length()-1)));
                continue;
            }
            // end old version
        }
        // if we can reach this condition, we have understand that it is an old server
        // and put smartSlowSync to 1.
	    if(getRegConfig()->getServerConfig().getSmartSlowSync() == 2) {
		    getRegConfig()->getServerConfig().setSmartSlowSync(1);
	    }
        getRegConfig()->save();
	}
}
